#ifndef UUGEOM38219HSNRUIO439578SANTIAA
#define UUGEOM38219HSNRUIO439578SANTIAA

// vim:set tabstop=4 cindent:

// Author: Geert-Jan Giezeman <geert@cs.uu.nl>
// Copyright: Utrecht University, 2011

#include <cassert>

namespace UuIcsGeom {

	typedef double DTriple[3];
	typedef float FTriple[3];

	//! \brief Class for 3 dimensional vectors (x,y,z) with doubles as coordinates.
	struct Vector3d {
		DTriple coords;
		Vector3d() {}
		Vector3d(double c0, double c1, double c2) { set(c0,c1,c2);}
		explicit Vector3d(DTriple const &in){set(in[0],in[1],in[2]);}
		explicit Vector3d(FTriple const &in){set(in[0],in[1],in[2]);}
		void set(double c0, double c1, double c2)
		{
			coords[0]=c0;
			coords[1]=c1;
			coords[2]=c2;
		}
		Vector3d operator-() const
		{
			return Vector3d(-coords[0], -coords[1],-coords[2]);
		}
		void operator+=(Vector3d const &o)
		{ 
			coords[0]+=o.coords[0];
			coords[1]+=o.coords[1];
			coords[2]+=o.coords[2];
		}
		void operator+=(DTriple const &o)
		{ 
			coords[0]+=o[0];
			coords[1]+=o[1];
			coords[2]+=o[2];
		}
		void operator-=(Vector3d const &o)
		{ 
			coords[0]-=o.coords[0];
			coords[1]-=o.coords[1];
			coords[2]-=o.coords[2];
		}
		void operator*=(double f)
		{ 
			coords[0]*=f;
			coords[1]*=f;
			coords[2]*=f;
		}
	};

	inline Vector3d operator+(Vector3d v1, Vector3d const &v2)
	{
		v1 += v2;
		return v1;
	}

	inline Vector3d operator-(Vector3d v1, Vector3d const &v2)
	{
		v1 -= v2;
		return v1;
	}

	double inprod(DTriple const &v1, DTriple const &v2);
	inline double inprod(Vector3d const &v1, Vector3d const &v2)
	{
		return inprod(v1.coords, v2.coords);
	}

	Vector3d cross(DTriple const &vec0, DTriple const &vec1);

	double vector_length(DTriple const &vec);

	Vector3d diffvec(DTriple const &pt1, DTriple const &pt2);

	inline void mult_by(DTriple &vec, double f)
	{
		for (int i=0; i!=3; ++i) {
			vec[i] *= f;
		}
	}

	inline Vector3d operator*(double f, Vector3d v)
	{
		mult_by(v.coords, f);
		return v;
	}

	//! A class representing points in 3d space, with doubles as coordinates.
	struct Point3d: Vector3d {
		Point3d() {}
		explicit Point3d(DTriple const &in):Vector3d(in) {}
		explicit Point3d(FTriple const &in):Vector3d(in) {}
		Point3d(double c0, double c1, double c2) {set(c0,c1,c2);}
		explicit Point3d(Vector3d const &in): Vector3d(in) {}
	};

	typedef double DNonet[9];

	//! Class reprsenting a 3 by 3 square matrix with doubles as values.
	struct Matrix3d {
		DNonet coords;
		Matrix3d()
		{}
		Matrix3d(double c0, double c1, double c2, double c3,double c4, double c5,
				double c6, double c7, double c8)
		{ set(c0,c1,c2,c3,c4,c5,c6,c7,c8);}
		explicit Matrix3d(DNonet const &in)
		{set(in[0],in[1],in[2],in[3],in[4],in[5],in[6],in[7],in[8]);}
		static Matrix3d const &identity();
		void set(double c0, double c1, double c2, double c3,double c4, double c5,
				double c6, double c7, double c8)
		{
			coords[0]=c0;
			coords[1]=c1;
			coords[2]=c2;
			coords[3]=c3;
			coords[4]=c4;
			coords[5]=c5;
			coords[6]=c6;
			coords[7]=c7;
			coords[8]=c8;
		}
		double value_raw(int i, int j) const
		{ return coords[3*i+j];}
		void set_value_raw(int i, int j, double v)
		{ coords[3*i+j] = v;}
		double operator()(int i, int j) const
		{ 
			assert(0<=i);
			assert(i<3);
			assert(0<=j);
			assert(j<3);
			return value_raw(i,j);
		}
		double set_value(int i, int j, double v)
		{ 
			assert(0<=i);
			assert(i<3);
			assert(0<=j);
			assert(j<3);
			set_value_raw(i,j,v);
		}
	};

	Vector3d mult(Matrix3d const &, DTriple const &);
	inline Vector3d mult(Matrix3d const &m, Vector3d const &v)
	{ return mult(m,v.coords);}
	inline Vector3d operator*(Matrix3d const &m, Vector3d const &v)
	{ return mult(m,v.coords);}

	Matrix3d matrixmult(DNonet const &, DNonet const &);
	inline Matrix3d mult(Matrix3d const &m1, DNonet const &m2)
	{ return matrixmult(m1.coords, m2); }
	inline Matrix3d mult(DNonet const &m1, Matrix3d const &m2)
	{ return matrixmult(m1, m2.coords); }
	inline Matrix3d mult(Matrix3d const &m1, Matrix3d const &m2)
	{ return matrixmult(m1.coords, m2.coords); }
	inline Matrix3d operator*(Matrix3d const &m1, Matrix3d const &m2)
	{ return matrixmult(m1.coords, m2.coords); }

	Matrix3d rotation_matrix_from_unit_axis(DTriple const &unit_axis, double rad);
	inline Matrix3d rotation_matrix_from_unit_axis(
			Vector3d const &unit_axis, double rad)
	{ return rotation_matrix_from_unit_axis(unit_axis.coords, rad); }
	Matrix3d rotation_matrix_from_axis(Vector3d axis, double rad);
	inline Matrix3d rotation_matrix_from_axis(DTriple const &axis, double rad)
	{ return rotation_matrix_from_axis(Vector3d(axis), rad); }


	//! \brief An orthonormal basis for 3d space
	//!
	struct Basis3d {
		Basis3d() {}
		Basis3d( Point3d const &origin, Vector3d const &axis1,
		Vector3d const &axis2, Vector3d const &axis3) :
				m_reference(origin){ m_axis[0]=axis1;
				m_axis[1]=axis2; m_axis[2]=axis3;}
		// orthonormal vectors
		Point3d reference_point() const
		{ return m_reference;}
		Vector3d axis(int i) const
		{ assert(i>=0); assert(i<3); return m_axis[i]; }
		//! Point pt expressed as a vector relative to this basis
		Vector3d relative_to_this(Point3d const &pt)
		{return relative_to_this(pt.coords);} 
		Vector3d relative_to_this(DTriple const &pt);
		//Vector3f relative_to_this(Point3f const &pt)
		//{return relative_to_this(pt.coords);}
		//Vector3f relative_to_this(FTriple const &pt);
		private:
		Point3d m_reference;
		Vector3d m_axis[3]; // orthonormal vectors
	};

		//! area of parallellogram of defining points.
		//! If 0, there was a degenerate case.
	Basis3d make_basis1(double &definition_area, DTriple const &pt1,
					DTriple const &pt2, DTriple const &pt3);
	inline Basis3d
		make_basis1(double &definition_area,
				Point3d const &pt1, Point3d const &pt2, Point3d const &pt3)
    {
		return make_basis1(definition_area, pt1.coords, pt2.coords, pt3.coords);
	}
    Basis3d make_basis1(double &def_area, FTriple const &pt1,
				FTriple const &pt2, FTriple const &pt3);

	inline Basis3d
		make_basis1(DTriple const &pt1, DTriple const &pt2, DTriple const &pt3)
	{
		double dummy;
		return make_basis1(dummy, pt1,pt2,pt3);
	}
	inline Basis3d
		make_basis1(Point3d const &pt1, Point3d const &pt2, Point3d const &pt3)
	{
		double dummy;
		return make_basis1(dummy, pt1.coords, pt2.coords, pt3.coords);
	}
    inline Basis3d
		make_basis1(FTriple const &pt1, FTriple const &pt2, FTriple const &pt3)
	{
		double dummy;
		return make_basis1(dummy, pt1, pt2,pt3);
	}

	void check_basis();
}

#endif // UUGEOM38219HSNRUIO439578SANTIAA
