// vim:set sw=4 ts=4 cindent:

// Author: Geert-Jan Giezeman <geert@cs.uu.nl>
// Copyright: Utrecht University, 2011

#include "label_anonymous.hpp"
#include "skeleton_info.hpp"
#include "uuc3d.hpp"
#include "subjects.hpp"
#include "geometry.hpp"
#include <queue>
#include <list>

#include <iostream>

using std::string;
using std::vector;
using std::list;
using std::priority_queue;
using UuIcsC3d::SpacePaddedString;
using UuIcsC3d::FrameData;
using UuIcsC3d::LabeledSkeleton;
using UuIcsC3d::DataPoint3d;
using ::UuIcsGeom::Point3d;
using ::UuIcsGeom::Vector3d;
	
namespace {

	struct PqLast {};
	struct PqStartUncertain{};
	struct PqEndUncertain{};
	struct PqStartAnonymous{};
	struct PqEndAnonymous{};

	enum PqType {pq_last=0, pq_start_uncertain, pq_end_uncertain,
		pq_start_anonymous, pq_end_anonymous};

	struct PqEvent {
			PqEvent(PqLast , int framecount)
					:frame_nr(framecount), prio(100), type(pq_last) {}
			PqEvent(PqStartUncertain, int frame_nr_in, int label)
					:frame_nr(frame_nr_in), prio(5),
					type(pq_start_uncertain), marker(label)
			{}
			PqEvent(PqEndUncertain , int frame_nr_in, int label)
					:frame_nr(frame_nr_in), prio(1),
					type(pq_end_uncertain), marker(label)
			{}
			PqEvent(PqStartAnonymous , int frame_nr_in, int lbl_in)
					:frame_nr(frame_nr_in), prio(3), type(pq_start_anonymous),
					marker(lbl_in)
			{}
			PqEvent(PqEndAnonymous , int frame_nr_in, int lbl_in, int start)
					:frame_nr(frame_nr_in), prio(3), type(pq_end_anonymous),
					marker(lbl_in), start_frame(start)
			{}
			int frame_nr;
			int prio;
			PqType type;
			//MarkerIndex mi;
			int marker;
			int start_frame; //!< start frame of anony_stretch
	};

	//! PqEvent are compared according to priority
	//! A later (higher) frame has less priority
	//! Events for the same frame are ordered according to the prio field
	bool operator<(PqEvent const &ev1, PqEvent const &ev2)
	{
		if (ev1.frame_nr < ev2.frame_nr)
			return false;
		else if (ev2.frame_nr < ev1.frame_nr) {
			return true;
		} else {
			return ev1.prio < ev2.prio;
		}
	}

	struct CandidateMatch {
		CandidateMatch(int a, int l, int ss, int se, int mf)
		   :anony_id(a), label_id(l), stretch_start(ss), stretch_end(se), match_frame(mf)
		{}	
		int anony_id;
		int label_id;
		int stretch_start; // frame where anonymous stretch starts
		int stretch_end; // frame where anonymous stretch ends
		int match_frame; // frame where match_was found
	};

	bool operator<(CandidateMatch const &c1, CandidateMatch const &c2)
	{
		if (c1.stretch_start<c2.stretch_start)
			return true;
		if (c1.stretch_start>c2.stretch_start)
			return false;
		if (c1.anony_id<c2.anony_id)
			return true;
		if (c1.anony_id>c2.anony_id)
			return false;
		return false;
	}
};

/*
Design: link anonymous markers to labels.
1) Gather statistics about maximum speed of a label (max_speed).
2) Group anonymous markers in stretches (close to each other, relative to the speed)
3) For each stretch, check whether the begin or end is close to a interpolated label position in that frame.
-  check whether the complete stretch can be fitted to one interpolated label path, that is
   no actual label on the stretch (would be a duplicate)
   begin and end of the stretch linked to the same label (or only one was mapped)
4) Replace interpolated positions by anonymous marker positions (removing the anonymous markers)
5) Redo interpolation where markers are inserted.
*/

namespace {

	void push_next_uncertain(priority_queue<PqEvent> &events, 
		vector<FrameData> const &frame_data,int framenr, int idx)
	{
		int const frame_count = frame_data.size();
		for ( ; framenr!= frame_count; ++framenr) {
			if (frame_data[framenr].points[idx].status() ==
				DataPoint3d::Generated) {
					events.push(PqEvent(
						PqStartUncertain(),framenr, idx));
					break;
			}
		}
	}

	void find_start_of_generated(priority_queue<PqEvent> &events,
		vector<LabeledSkeleton> const &subjects, 
		vector<FrameData> const &frame_data)
	{
//		MarkerIndex mi;
		for (size_t s=0; s!= subjects.size(); ++s) {
			vector< vector<int> > const &cur_point_index =
				subjects[s].point_index();
			for (size_t g=0; g!= cur_point_index.size(); ++g) {
				vector<int> const &mem_idx(cur_point_index[g]);
				for ( size_t m=0; m!=mem_idx.size(); ++m) {
					int idx = mem_idx[m];
					if (idx>=0) {
						push_next_uncertain(events, frame_data, 0, idx);
					}
				}
			}
		}
	}

	void push_next_start( priority_queue<PqEvent> &events,
		vector<FrameData> &frame_data, int framenr, int idx)
	{
		int const frame_count = frame_data.size();
		for (; framenr!= frame_count; ++framenr) {
			if (frame_data[framenr].points[idx].is_valid()) {
				events.push(PqEvent(PqStartAnonymous(),framenr, idx));
				break;
			}
		}
	}

	void push_final_uncertain(priority_queue<PqEvent> &events,
		vector<FrameData> &frame_data, int frame, int idx)
	{
		int const frame_count = frame_data.size();
		assert(frame_data[frame].points[idx].status() ==DataPoint3d::Generated);
		for (frame = frame+1; frame != frame_count; ++frame) {
			if(!frame_data[frame].points[idx].status() != DataPoint3d::Generated)
				break;
		}
		events.push(PqEvent(PqEndUncertain(), frame-1, idx));
	}

	int find_end_of_anonymous_stretch(vector<FrameData> &frame_data,
		int idx, int frame, double max_speed)
	{
		int const frame_count = frame_data.size();
		assert(frame_data[frame].points[idx].is_valid());
		double cutoff = max_speed*max_speed;
		Point3d prev(frame_data[frame].points[idx].coordinates());
		for (frame = frame+1; frame != frame_count; ++frame) {
			if(!frame_data[frame].points[idx].is_valid())
				break;
			Point3d cur(frame_data[frame].points[idx].coordinates());
			Vector3d diff = cur-prev;
			if (inprod(diff,diff)>cutoff)
				break;
		}
		return frame-1;
	}

	void find_candidates(vector<CandidateMatch> &candidates,
		vector<FrameData> &frame_data, list<int> const &uncertain,
		int frame, int start, int end, int marker, double cutoff)
	{
		Point3d anony(frame_data[frame].points[marker].coordinates());
		for (list<int>::const_iterator cur=uncertain.begin();
			cur != uncertain.end(); ++cur) {
				Point3d interpolation_pt(
						frame_data[frame].points[*cur].coordinates());
				Vector3d diff = anony - interpolation_pt;
				if (inprod(diff,diff)<=cutoff) {
					candidates.push_back(CandidateMatch(marker, *cur, start, end, frame));
				}

		}
	}

}


//! compute anonymous label stretches
vector<UuIcsC3d::LabeledStretch> UuIcsC3d::label_anonymous_markers(
				vector<FrameData> &frame_data,
				vector<LabeledSkeleton> const &subjects, 
				vector<unsigned> const &no_label,
				double max_speed,
				vector<SpacePaddedString> const &labels)
{
	//! Cutoff value for when (squared distance of) points are close 
	double const MaxD = 25.0*25.0;
	int const frame_count = frame_data.size();

	// maintain a list of events:
	priority_queue<PqEvent> events;
	// find start of a anonymous marker stretch
	for (size_t i=0; i!=no_label.size(); ++i) {
		push_next_start(events, frame_data, 0, no_label[i]);
	}
	find_start_of_generated(events, subjects, frame_data);
	// maintain variable whether anonymous stretch is being inserted (and which).
	events.push(PqEvent(PqLast(),frame_count));
	//int cur_frame = -1;
	list<int> uncertain;
	vector<CandidateMatch> candidates;
	while (!events.empty()) {
		PqEvent ev = events.top();
		events.pop();
		switch( ev.type) {
		case pq_last:
			break;
		case pq_start_uncertain:
			uncertain.push_back(ev.marker);
			// find end of stretch and insert event
			push_final_uncertain(events,frame_data, ev.frame_nr,
				ev.marker);	
			break;
		case pq_end_uncertain:
			// remove from inserters
			uncertain.remove(ev.marker);
			// find next start of stretch and insert event
			push_next_uncertain(events, frame_data, ev.frame_nr+1, ev.marker);
			break;
		case pq_start_anonymous:
			{
				// Check if this point is close to an uncertain point
				int end = find_end_of_anonymous_stretch(frame_data,
						ev.marker, ev.frame_nr, max_speed);
				find_candidates(candidates, frame_data, uncertain,
						ev.frame_nr, ev.frame_nr, end, ev.marker, MaxD);
				if (end==ev.frame_nr) {
					push_next_start(events, frame_data,
							ev.frame_nr+1, ev.marker);
				} else {
					events.push(PqEvent(PqEndAnonymous(),
								end, ev.marker, ev.frame_nr));
				}
			}
		   break;
		case pq_end_anonymous:
			// Check if this point is close to an uncertain point
			find_candidates(candidates, frame_data, uncertain,
						ev.frame_nr, ev.start_frame, ev.frame_nr, ev.marker, MaxD);
			push_next_start(events, frame_data,
						ev.frame_nr+1, ev.marker);
		   break;
		}
	}
	std::sort(candidates.begin(), candidates.end());
	vector<LabeledStretch> result;
	int last_ano = -1;
	int last_start = -1;
	for (size_t i=0; i<candidates.size(); ++i) {
		CandidateMatch const &cand(candidates[i]);
		if (cand.anony_id == last_ano && cand.stretch_start == last_start)
			continue;
		bool all_uncertain = true;
		for (int j=cand.stretch_start; j<=cand.stretch_end; ++j) {
			if (frame_data[j].points[cand.label_id].status() == DataPoint3d::Measured) {
				all_uncertain = false;
				break;
			}
		}
		if (all_uncertain) {
			last_ano= cand.anony_id;
			last_start = cand.stretch_start;
			std::cout<<"Labeled: "<<labels[cand.label_id].stripped()<<' '
				<<cand.anony_id<<' '
				<<cand.stretch_start<<' '<<cand.stretch_end<<"\n";
			for (int j=cand.stretch_start; j<=cand.stretch_end; ++j) {
				DataPoint3d &anony_pt(frame_data[j].points[cand.anony_id]);
				frame_data[j].points[cand.label_id].set_value(
						anony_pt.coordinates(), 0, DataPoint3d::Measured);
				anony_pt.invalidate();
			}
			result.push_back(LabeledStretch(cand.label_id, cand.stretch_start, cand.stretch_end));
		}
	}
	return result;
}
