// vim:set tabstop=4 cindent tw=80:

// Author: Geert-Jan Giezeman <geert@cs.uu.nl>
// Copyright: Utrecht University, 2011

#include "uuc3d.hpp"
#include <string>
#include <cstdio>
#include "basic_io.hpp"
#include <boost/shared_ptr.hpp>
#include <cassert>
#include <map>

//#include <iostream>  // for debugging

using boost::shared_ptr;
using std::vector;
using std::string;
using UuIcsC3d::C3dFileInfo;
using UuIcsC3d::BasicIO;
using UuIcsC3d::unsigned16_t;

namespace {

	unsigned const BlockSize = 512;  // number of bytes in a block

	shared_ptr<FILE> open_c_file(char const *filename, char const*flags="rb")
	{
		FILE *fp=fopen(filename,flags);
		if (fp==0) {
			throw UuIcsC3d::OpenError(filename);
		}
		shared_ptr<FILE> fptr(fp, fclose);
		return fptr;
	}

	std::string read_name1(unsigned char const *buf_start, int n,
					unsigned char const *buf_end)
	{
		if (buf_start+n >buf_end)
			throw UuIcsC3d::ContentError("String too large to fit in buffer","");
		char const*name_start = reinterpret_cast<char const*>(buf_start);
		return string(name_start, n);
	}

	string get_stripped_string(unsigned char const *str, int sz)
	{
		vector<char> str2(sz+1,0);
		std::copy(str,str+sz,str2.begin());
		return string(&str2[0]);
	}

	struct HasId
	{
		HasId(int id):m_id(id) {}
		bool operator()(UuIcsC3d::Group const &group) const
		{ return group.id() == m_id; }
	private:
		int m_id;
	};

	struct HasName {
		HasName(string const &name):m_name(name) {}
		bool operator()(UuIcsC3d::Parameter const &p)
		{
			return p.name() == m_name;
		}
	  private:
		string m_name;
	};
}

// **** Group ****


void UuIcsC3d::Group::add_parameter(Parameter const &param)
{
    m_parameters.push_back(param);
}

bool UuIcsC3d::Group::has_parameter(string const &name) const
{
    return find_if(m_parameters.begin(), m_parameters.end(), HasName(name))
			!= m_parameters.end();
}

int UuIcsC3d::Group::group_representation_size() const
{
	return 5 + this->m_name.size() + description().size();
}

int UuIcsC3d::Group::total_representation_size() const
{
	int result = group_representation_size();
	for (vector<Parameter>::const_iterator cur=m_parameters.begin();
					cur != m_parameters.end(); ++cur) {
		result += cur->representation_size();
	}
	return result;
}


UuIcsC3d::Parameter const
&UuIcsC3d::Group::get_parameter(string const &name) const
{
    vector<Parameter>::const_iterator it= 
			find_if(m_parameters.begin(), m_parameters.end(), HasName(name));
	if (it == m_parameters.end())
    	throw PreconditionError();
	return *it;
}

UuIcsC3d::Parameter &UuIcsC3d::Group::get_parameter(string const &name)
{
    vector<Parameter>::iterator it= 
			find_if(m_parameters.begin(), m_parameters.end(), HasName(name));
	if (it == m_parameters.end())
    	throw PreconditionError();
	return *it;
}

UuIcsC3d::Parameter const *UuIcsC3d::Group::get_parameter_checked(
				string const &name) const
{
    vector<Parameter>::const_iterator it= 
			find_if(m_parameters.begin(), m_parameters.end(), HasName(name));
	if (it == m_parameters.end())
    	return 0;
	return &*it;
}

UuIcsC3d::Parameter *UuIcsC3d::Group::get_parameter_checked(
				string const &name)
{
    vector<Parameter>::iterator it= 
			find_if(m_parameters.begin(), m_parameters.end(), HasName(name));
	if (it == m_parameters.end())
    	return 0;
	return &*it;
}

// **** Parameter ****

UuIcsC3d::Parameter::Parameter(bool locked, int id, std::string name,
		int data_type, std::vector<int> const &dimension_size,
		unsigned char const*data_start, unsigned char const*data_end,
		std::string const &description,
		boost::shared_ptr<BasicIO> io)
            : GroupParamCommon(locked, id, name, description), m_dimension_size(dimension_size), 
			 m_data(data_start, data_end), m_io(io)
{
    assert(data_type==-1||data_type==1||data_type==2||data_type==4);
    if (data_type==-1||data_type==1||data_type==2||data_type==4)
        m_data_type = static_cast<DataType>(data_type);
    else
        throw "Illegal Value";
}

int UuIcsC3d::Parameter::representation_size() const
{
	return 7 + this->m_name.size() + m_description.size()
			+ m_data.size() + m_dimension_size.size();
}


bool UuIcsC3d::Parameter::check_index(std::vector<int> const &index) const
{
    if (index.size() != m_dimension_size.size())
        return false;
    for (size_t i=0; i<index.size(); ++i) {
        if (index[i]<0 || m_dimension_size[i]<=index[i])
            return false;
    };
    return true;
}

bool UuIcsC3d::Parameter::check_string_index(std::vector<int> const &index) const
{
    if (index.size()+1 != m_dimension_size.size())
        return false;
    for (size_t i=0; i<index.size(); ++i) {
        if (index[i]<0 || m_dimension_size[i]<=index[i])
            return false;
    };
    return true;
}

int UuIcsC3d::Parameter::get_offset(int data_size,
				std::vector<int> const &index) const
{
    int offset = 0;
    int fac = data_size;
    for (size_t i=index.size(); i>0; --i) {
        offset += fac*index[i-1];
        fac *= m_dimension_size[i-1];
    }
    return offset;
}

int UuIcsC3d::Parameter::get_string_offset(std::vector<int> const &index) const
{
	assert(index.size()+1==m_dimension_size.size());
    int offset = 0;
    int fac = m_dimension_size.back();
    for (size_t i=index.size(); i>0; --i) {
        offset += fac*index[i-1];
        fac *= m_dimension_size[i-1];
    }
    return offset;
}

unsigned char UuIcsC3d::Parameter::get_byte(std::vector<int> const &index) const
{
    if (m_data_type != DtByte)
        throw PreconditionError();
    if (!check_index(index))
        throw PreconditionError();
    int offset = get_offset(1, index);
    return m_data[offset];
}

unsigned char UuIcsC3d::Parameter::get_byte(int i1) const
{
    if (m_data_type != DtByte)
        throw PreconditionError();
	if (m_dimension_size.size() != 1)
		throw PreconditionError();
	if (i1<0 || m_dimension_size[0] <= i1)
		throw PreconditionError();
	return m_data[i1];
}

unsigned char UuIcsC3d::Parameter::get_the_byte() const
{
    if (m_data_type != DtByte)
        throw PreconditionError();
	for (size_t i=0; i<m_dimension_size.size(); ++i) {
		if (m_dimension_size[i] != 1)
			throw PreconditionError();
	}
	return m_data[0];
}

char UuIcsC3d::Parameter::get_char(std::vector<int> const &index) const
{
    if (m_data_type != DtChar)
        throw PreconditionError();
    if (!check_index(index))
        throw PreconditionError();
    int offset = get_offset(1, index);
    return m_data[offset];
}

char UuIcsC3d::Parameter::get_char(int i1) const
{
    if (m_data_type != DtChar)
        throw PreconditionError();
	if (m_dimension_size.size() != 1)
		throw PreconditionError();
	if (i1<0 || m_dimension_size[0] <= i1)
		throw PreconditionError();
	return m_data[i1];
}

char UuIcsC3d::Parameter::get_the_char() const
{
    if (m_data_type != DtChar)
        throw PreconditionError();
	for (size_t i=0; i<m_dimension_size.size(); ++i) {
		if (m_dimension_size[i] != 1)
			throw PreconditionError();
	}
	return m_data[0];
}

bool UuIcsC3d::Parameter::contains_string() const
{
    if (m_data_type != DtChar)
        return false;
	if (m_dimension_size.size()==0)
		return false;
	return true;
}

std::string UuIcsC3d::Parameter::get_string(std::vector<int> const &index) const
{
    if (m_data_type != DtChar)
        throw PreconditionError();
    if (!check_string_index(index))
        throw PreconditionError();
    int offset = get_string_offset(index);
	return get_stripped_string(&m_data[0]+offset, m_dimension_size.back());
}

std::string UuIcsC3d::Parameter::get_string(int i1) const
{
    if (m_data_type != DtChar)
        throw PreconditionError();
	if (m_dimension_size.size() != 2)
		throw PreconditionError();
	if (i1<0 || m_dimension_size[0] <= i1)
		throw PreconditionError();
	int str_sz = m_dimension_size.back();
	return get_stripped_string(&m_data[0]+i1*str_sz, str_sz);
}


std::string UuIcsC3d::Parameter::get_the_string() const
{
    if (!contains_string())
        throw PreconditionError();
	for (size_t i=0; i<m_dimension_size.size()-1; ++i) {
		if (m_dimension_size[i] != 1)
			throw PreconditionError();
	}
	return get_stripped_string(&m_data[0], m_dimension_size.back());
}


int UuIcsC3d::Parameter::get_short(std::vector<int> const &index) const
{
    if (m_data_type != DtShort)
        throw PreconditionError();
    if (!check_index(index))
        throw PreconditionError();
    int offset = get_offset(2, index);
    return m_io->to_int16(&m_data[0]+offset);
}

int UuIcsC3d::Parameter::get_short(int i1) const
{
    if (m_data_type != DtShort)
        throw PreconditionError();
	if (m_dimension_size.size() != 1)
		throw PreconditionError();
	if (i1<0 || m_dimension_size[0] <= i1)
		throw PreconditionError();
	return m_io->to_int16(&m_data[0]+2*i1);
}

int UuIcsC3d::Parameter::get_the_short() const
{
    if (m_data_type != DtShort)
        throw PreconditionError();
	for (size_t i=0; i<m_dimension_size.size(); ++i) {
		if (m_dimension_size[i] != 1)
			throw PreconditionError();
	}
	return m_io->to_int16(&m_data[0]);
}

unsigned UuIcsC3d::Parameter::get_ushort(std::vector<int> const &index) const
{
    if (m_data_type != DtShort)
        throw PreconditionError();
    if (!check_index(index))
        throw PreconditionError();
    int offset = get_offset(2, index);
    return m_io->to_uint16(&m_data[0]+offset);
}

unsigned UuIcsC3d::Parameter::get_ushort(int i1) const
{
    if (m_data_type != DtShort)
        throw PreconditionError();
	if (m_dimension_size.size() != 1)
		throw PreconditionError();
	if (i1<0 || m_dimension_size[0] <= i1)
		throw PreconditionError();
	return m_io->to_uint16(&m_data[0]+2*i1);
}

unsigned UuIcsC3d::Parameter::get_the_ushort() const
{
    if (m_data_type != DtShort)
        throw PreconditionError();
	for (size_t i=0; i<m_dimension_size.size(); ++i) {
		if (m_dimension_size[i] != 1)
			throw PreconditionError();
	}
	return m_io->to_uint16(&m_data[0]);
}

void UuIcsC3d::Parameter::set_the_ushort(unsigned16_t val)
{
    if (m_data_type != DtShort)
        throw PreconditionError();
	for (size_t i=0; i<m_dimension_size.size(); ++i) {
		if (m_dimension_size[i] != 1)
			throw PreconditionError();
	}
	m_io->encode(&m_data[0], val);
}

void UuIcsC3d::Parameter::set_ushort(int i1, unsigned16_t val)
{
    if (m_data_type != DtShort)
        throw PreconditionError();
	if (m_dimension_size.size() != 1)
		throw PreconditionError();
	if (i1<0 || m_dimension_size[0] <= i1)
		throw PreconditionError();
	m_io->encode(&m_data[0]+2*i1, val);
}

void UuIcsC3d::Parameter::set_strings(vector<string> const &data, char pad)
{
	typedef vector<string>::const_iterator ci_t;
    //if (is_locked())
    //    throw PreconditionError();
    if (m_data_type != DtChar)
        throw PreconditionError();
	unsigned string_count = data.size();
	unsigned string_size = string_count==0;
	ci_t end = data.end();
	for (ci_t cur = data.begin(); cur != end; ++cur) {
		unsigned sz = cur->size();
		if (sz>string_size)
			string_size=sz;
	}
	m_dimension_size.resize(2);
	m_dimension_size[0] = string_count;
	m_dimension_size[1] = string_size;
	m_data.resize(string_count*string_size);
	vector<unsigned char>::iterator it = m_data.begin();
	for (ci_t cur = data.begin(); cur != end; ++cur) {
		unsigned p = 0;
		while (p<cur->size()) {
			*it++ = (*cur)[p++];
		}
		while (p<string_size) {
			*it++ = pad;
			++p;
		}
	}
}

float UuIcsC3d::Parameter::get_float(std::vector<int> const &index) const
{
    if (m_data_type != DtFloat)
        throw PreconditionError();
    if (!check_index(index))
        throw PreconditionError();
    int offset = get_offset(4, index);
    return m_io->to_float(&m_data[0]+offset);
}

float UuIcsC3d::Parameter::get_float(int i1) const
{
    if (m_data_type != DtShort)
        throw PreconditionError();
	if (m_dimension_size.size() != 1)
		throw PreconditionError();
	if (i1<0 || m_dimension_size[0] <= i1)
		throw PreconditionError();
	return m_io->to_float(&m_data[0]+4*i1);
}

float UuIcsC3d::Parameter::get_the_float() const
{
    if (m_data_type != DtShort)
        throw PreconditionError();
	for (size_t i=0; i<m_dimension_size.size(); ++i) {
		if (m_dimension_size[i] != 1)
			throw PreconditionError();
	}
	return m_io->to_float(&m_data[0]);
}

// **** Content ****
namespace {

	void parse_header_block(UuIcsC3d::Header &header,
					UuIcsC3d::Block header_block, BasicIO *io)
	{
		header.trajectory_count = io->to_uint16(header_block+2);
		header.analog_per_frame = io->to_uint16(header_block+4);
		header.first_frame = io->to_uint16(header_block+6);
		header.last_frame = io->to_uint16(header_block+8);
		header.max_interpolation_gap = io->to_uint16(header_block+10);
		header.scaling_factor = io->to_float(header_block+12);
		header.data_start_block = io->to_uint16(header_block+16);
		header.analog_samples_per_frame = io->to_uint16(header_block+18);
		header.frame_rate = io->to_float(header_block+20);
		header.has_label_and_range = (io->to_uint16(header_block+147*2) == 12345);
		header.label_and_range_block = io->to_uint16(header_block+148*2);
		int event_count = io->to_uint16(header_block+150*2);
		header.events.resize(event_count);
		for (int ev=0; ev != event_count; ++ev) {
			UuIcsC3d::HeaderEvent &cur_event = header.events[ev];
			cur_event.time = io->to_float(header_block+142*2+4*ev);
			cur_event.displayed = (header_block[188*2+ev] == 0);
			unsigned char * label_start = header_block+198*2+ev;
			cur_event.label = string(label_start, label_start+4);
		}
	}

	void write_header_block(UuIcsC3d::Block &header_block,
			UuIcsC3d::Header const &header, BasicIO const*io)
	{
		std::fill_n(header_block, BlockSize, 0);
		unsigned char *hb = header_block;
		hb[0]  = 2;
		hb[1] = 0x50;
		io->encode(hb+2, static_cast<unsigned16_t>(header.trajectory_count));
		io->encode(hb+4, static_cast<unsigned16_t>(header.analog_per_frame));
		io->encode(hb+6, static_cast<unsigned16_t>(header.first_frame));
		io->encode(hb+8, static_cast<unsigned16_t>(header.last_frame));
		io->encode(hb+10, static_cast<unsigned16_t>(header.max_interpolation_gap));
		io->encode(hb+12, header.scaling_factor);
		io->encode(hb+16, static_cast<unsigned16_t>(header.data_start_block));
		io->encode(hb+18, static_cast<unsigned16_t>(header.analog_samples_per_frame));
		io->encode(hb+20, header.frame_rate);
		io->encode(hb+147*2, static_cast<unsigned16_t>(header.has_label_and_range?12345u:0u));
		io->encode(hb+148*2, static_cast<unsigned16_t>(header.label_and_range_block));
		io->encode(hb+149*2, static_cast<unsigned16_t>(0x3039u));
		int event_count = header.events.size();
		io->encode(hb+150*2, static_cast<unsigned16_t>(header.events.size()));
		for (int ev=0; ev != event_count; ++ev) {
			UuIcsC3d::HeaderEvent const&cur_event = header.events[ev];
			io->encode(hb+142*2+4*ev, cur_event.time);
			header_block[188*2+ev] = cur_event.displayed ? 0 : 1;
			unsigned char * label_start = hb+198*2+ev;
			std::copy(cur_event.label.begin(), cur_event.label.end(), label_start);
		}
	}

}

UuIcsC3d::Content::Content()
{}

UuIcsC3d::Content UuIcsC3d::open_c3d_file(char const *filename)
{
    UuIcsC3d::Content result;
    result.open(filename);
    return result;
}

UuIcsC3d::Content UuIcsC3d::open_c3d_file(std::string const &filename)
{
    UuIcsC3d::Content result;
    result.open(filename);
    return result;
}

void UuIcsC3d::Content::open(std::string const &filename)
{
    shared_ptr<FILE> fptr(open_c_file(filename.c_str()));   
    m_filename = filename;
    FILE *fp = fptr.get();
	Block header_block;
    if (fread(header_block, 1, BlockSize, fp) != BlockSize)
        throw ContentError("Could not read complete header block", filename);
    if (header_block[1] != 0x50)
        throw  ContentError("Illegal byte value in header byte 2", filename);
	vector<unsigned char> parameter_section;
    int blockno = header_block[0]-1;
    if (blockno >1) {
        if (fseek(fp, (blockno-1)*BlockSize, SEEK_CUR)!=0)
            throw  ContentError("Could not seek to parameter section", filename);
    }
    Block param_block;
    if (fread(param_block, 1, BlockSize, fp) != BlockSize)
        throw ContentError("Could not read first parameter block", filename);
    int block_count = param_block[2];
    unsigned char processor_type = param_block[3];
    switch (processor_type) 
    {
        case Pt_Intel:
        case Pt_Dec:
        case Pt_Mips:
            //m_data_rep->m_io = get_basic_io(processor_type);
			m_io = get_basic_io(processor_type);
            break;
        default:
            throw ContentError("Unknown processor type", filename);
    }
	parse_header_block(header(), header_block, m_io.get());
    parameter_section.insert(parameter_section.end(),
					param_block, param_block+BlockSize);
    for (int i=1; i<block_count; ++i) {
        if (fread(param_block, 1, BlockSize, fp) != BlockSize)
            throw ContentError("Could not read parameter block", filename);
        parameter_section.insert(parameter_section.end(),
						param_block, param_block+BlockSize);
    }
    parse_params(parameter_section);
}


void UuIcsC3d::Content::parse_params(vector<unsigned char> const &parameter_section)
{
    size_t ps_size = parameter_section.size();
    assert(ps_size>0);
    unsigned char const *ps_beyond = &parameter_section[0]+ps_size;
    size_t cur_item_start = 4;
    while (cur_item_start<ps_size) {
        unsigned char const*cur = &parameter_section[cur_item_start];
        bool locked = false;
        signed char name_size = *cur++;
        if (name_size==0) {
            ++cur_item_start;
            continue;
        }
        if (name_size<0) {
            locked = true;
            name_size = -name_size;
        }
        int id = static_cast<signed char>(*cur++);
        std::string name(read_name1(cur, name_size, ps_beyond));
        cur += name_size;
        int offset = m_io->to_int16(cur);
        cur +=2;
        if (offset<=0) {
            throw ContentError("Negative offset in group/parameter",m_filename);
        }
		int total_size = offset+name_size+2;
        if (id<0) {
            int descr_size = *cur++;
            std::string description(read_name1(cur, descr_size, ps_beyond));
            vector<Group>::iterator group_it =
					std::find_if(m_groups.begin(), m_groups.end(), HasId(-id));
            if (group_it == m_groups.end()) {
                Group new_group(locked, -id, name, /*cur_item_start,*/ description);
                m_groups.push_back(new_group);
            } else {
                group_it->set_name(name);
				group_it->set_locked(locked);
                group_it->set_description(description);
            }
        } else {
            // read parameter data
            signed char data_type = *cur++;
            if (!(data_type==-1||data_type==1||data_type==2||data_type==4))
                throw ContentError("Illegal data type value in parameter",
								m_filename);
            int dimension_count = *cur++;
            if (dimension_count>7)
                throw ContentError(
						"Parameter with more than 7 dimensions encountered",
						m_filename);
            vector<int> dimension_size(dimension_count, -1);
            int data_size = abs(data_type);
            for (int i=0; i<dimension_count; ++i) {
                data_size *= *cur;
				// In file the Fortran array convention is used:
				//(fastest changing index in memory first)
				// We use the C convention (fastest changing in memory last)
                dimension_size[dimension_count -i-1] = *cur++;
            }
            unsigned char const*parameter_data_start = cur;
            cur += data_size;
            int descr_size = *cur++;
            string description(read_name1(cur, descr_size, ps_beyond));
			// TODO: convert data to native representation if necessary
			//(right endianness, floating point format)
			// until then, assert
			assert(this->io()->get_encoding() == platform_encoding());

            Parameter new_parameter(locked, id, name, /*cur_item_start, */
                data_type, dimension_size, parameter_data_start,
				parameter_data_start+data_size, description, m_io);
            vector<Group>::iterator group_it =
					std::find_if(m_groups.begin(), m_groups.end(), HasId(id));
            if (group_it == m_groups.end()) {
                Group new_group(id);
                new_group.add_parameter(new_parameter);
                m_groups.push_back(new_group);
            } else {
                group_it->add_parameter(new_parameter);
            }
        }
        cur_item_start = cur_item_start+total_size;
    }
}

bool UuIcsC3d::Content::has_group(string const &name) const
{
    for (vector<Group>::const_iterator cur=m_groups.begin();
					cur != m_groups.end(); ++cur)
        if (cur->name() == name)
            return true;
    return false;
}

UuIcsC3d::Group const &UuIcsC3d::Content::get_group(std::string const &name) const
{
    for (vector<Group>::const_iterator cur=m_groups.begin();
					cur != m_groups.end(); ++cur)
        if (cur->name() == name)
            return *cur;
    throw PreconditionError();
}

UuIcsC3d::Group & UuIcsC3d::Content::get_group(std::string const &name)
{
    for (vector<Group>::iterator cur=m_groups.begin();
					cur != m_groups.end(); ++cur)
        if (cur->name() == name)
            return *cur;
    throw PreconditionError();
}


UuIcsC3d::Group const*UuIcsC3d::Content::get_group_checked(std::string const &name) const
{
    for (vector<Group>::const_iterator cur=m_groups.begin();
					cur != m_groups.end(); ++cur)
        if (cur->name() == name) {
            return &*cur;
        }
    return 0;
}

UuIcsC3d::Group *UuIcsC3d::Content::get_group_checked(std::string const &name)
{
    for (vector<Group>::iterator cur=m_groups.begin();
					cur != m_groups.end(); ++cur)
        if (cur->name() == name) {
            return &*cur;
        }
    return 0;
}
// **** DataPoint3d ****

void UuIcsC3d::DataPoint3d::set_value(float x, float y, float z,
				unsigned gm, Status v)
{
	m_coords[0] = x;
	m_coords[1] = y;
	m_coords[2] = z;
	m_flags = gm;
	m_type = v;
}

void UuIcsC3d::DataPoint3d::set_value(float const*in_coordinates,
				unsigned gm, Status v)
{
	m_coords[0] = in_coordinates[0];
	m_coords[1] = in_coordinates[1];
	m_coords[2] = in_coordinates[2];
	m_flags = gm;
	m_type = v;
}

void UuIcsC3d::DataPoint3d::set_value(double const *in_coordinates,
				unsigned gm, Status v)
{
	m_coords[0] = static_cast<float>(in_coordinates[0]);
	m_coords[1] = static_cast<float>(in_coordinates[1]);
	m_coords[2] = static_cast<float>(in_coordinates[2]);
	m_flags = gm;
	m_type = v;
}

// **** SpacePaddedString ****

void UuIcsC3d::SpacePaddedString::compute_size()
{
	string::size_type pos=m_val.find_last_not_of(' ');
	m_size= pos==string::npos ? 0: pos+1;
}

// **** C3dFileInfo ****

C3dFileInfo::C3dFileInfo()
{}

C3dFileInfo::C3dFileInfo(string const &filename)
{
	m_content.open(filename);
	boost::shared_ptr<BasicIO> biop=m_content.io();
    UuIcsC3d::Header const &hd = m_content.header();
	Group const *point_group;
	Group const *analog_group;
	Group const *trial_group = m_content.get_group_checked("TRIAL");
	if ((point_group=m_content.get_group_checked("POINT"))==0) {
		throw ContentError("Required group POINT is missing", filename);
	}
	if ((analog_group=m_content.get_group_checked("ANALOG"))==0) {
		throw ContentError("Required group ANALOG is missing", filename);
	}
	Parameter const *param;
	if ((param=point_group->get_parameter_checked("FRAMES"))==0)
		throw ContentError("Required parameter POINT:FRAMES is missing", filename);
	m_fi1.frame_count = param->get_the_ushort();
	if (trial_group!=0) {
		// use ACTUAL_END_FIELD etc. to compute m_frame_count value
		unsigned asf=hd.first_frame, aef;
		if ((param=trial_group->get_parameter_checked("ACTUAL_START_FIELD"))!=0) {
			asf = param->get_ushort(1);
			asf <<=16;
			asf += param->get_ushort(0);
		}
		if ((param=trial_group->get_parameter_checked("ACTUAL_END_FIELD"))!=0) {
			aef = param->get_ushort(1);
			aef <<=16;
			aef += param->get_ushort(0);
			m_fi1.frame_count = aef-asf+1;
		}
	}
	if ((param=point_group->get_parameter_checked("DATA_START"))==0)
		throw ContentError("Required parameter POINT:DATA_START is missing",
						filename);
	m_fi1.data_startblock0 = param->get_the_ushort()-1;
	if ((param=point_group->get_parameter_checked("USED"))==0)
		throw ContentError("Required parameter POINT:USED is missing", filename);
	m_fi1.points_per_frame = param->get_the_ushort();
	m_fi1.samples_per_frame = hd.analog_samples_per_frame;
	int analog_per_frame_count = hd.analog_per_frame;
	//Bug, found and fixed by Dries Mys
	//m_fi1.analog_channels = analog_per_frame_count/m_fi1.points_per_frame; 
	m_fi1.analog_channels = analog_per_frame_count/m_fi1.samples_per_frame;

	m_fi1.point_scale = hd.scaling_factor;
	m_fi1.is_integer = true;
	if (m_fi1.point_scale <0) {
		m_fi1.point_scale = -m_fi1.point_scale;
		m_fi1.is_integer = false;
	}
	m_fi1.frame_size = (m_fi1.is_integer?2:4)
			*(analog_per_frame_count+4*m_fi1.points_per_frame);
	if ((param=point_group->get_parameter_checked("LABELS"))!=0) {
		std::vector<int> pb = param->bounds();
		if (pb.size()==2 && param->contains_string()) {
			std::vector<int> idx(2,0);
			for (int i=0; i<pb[0]; ++i) {
				m_labels.push_back(param->get_string(i));
			}
		}
	}
}

bool C3dFileInfo::is_valid() const
{
	return !m_content.filename().empty();
}

void C3dFileInfo::set_param_blocks(int c)
{
	if (c<0)
		throw PreconditionError();
	Header &hd(m_content.header());
	hd.data_start_block = c+2;
	Group *group;
	if ((group=m_content.get_group_checked("POINT"))!=0) {
		Parameter *param = group->get_parameter_checked("DATA_START");
		unsigned asf = hd.first_frame;
		if (param!=0) {
			param->set_the_ushort(static_cast<unsigned16_t>(c+2));
		}
	}
}
void C3dFileInfo::set_frame_count(int fc)
{
	if (fc<0)
		throw PreconditionError();
	m_fi1.frame_count = fc;
	Header &hd(m_content.header());
	unsigned16_t lf = hd.first_frame + static_cast<unsigned16_t>(fc)-1;
	hd.last_frame = lf;
	Group *group = m_content.get_group_checked("POINT");
	if (group!=0) {
		Parameter *param=group->get_parameter_checked("FRAMES");
		if (param!=0) {
			if (param->data_type() == Parameter::DtShort) {
				param->set_the_ushort(fc);
			}
		}
	}
	if ((group=m_content.get_group_checked("TRIAL"))!=0) {
		Parameter *param = group->get_parameter_checked("ACTUAL_END_FIELD");
		unsigned asf = hd.first_frame;
		if (param!=0) {
			unsigned aef = asf+fc-1;
			param->set_ushort(0,static_cast<unsigned16_t>(aef&0xffff));
			aef = aef>>16;
			param->set_ushort(1,static_cast<unsigned16_t>(aef&0xffff));
		}
	}
}

void C3dFileInfo::set_points_per_frame(int ppf)
{
	m_fi1.points_per_frame = ppf;
	Header &hd(m_content.header());
	unsigned16_t ppf2 = static_cast<unsigned16_t>(ppf);
	hd.trajectory_count=ppf2;
	Group *group = m_content.get_group_checked("POINT");
	if (group!=0) {
		Parameter *param=group->get_parameter_checked("USED");
		if (param!=0) {
			if (param->data_type() == Parameter::DtShort) {
				param->set_the_ushort(ppf);
			}
		}
	}

}

void C3dFileInfo::set_point_labels(vector<string> const &labels, char pad)
{
	// Find group and parameter
	Group *point_group = m_content.get_group_checked("POINT");
	if (point_group==0)
		throw NotImplemented();
	Parameter *labels_param = point_group->get_parameter_checked("LABELS");	
	if (labels_param==0)
		throw NotImplemented();
	labels_param->set_strings(labels, pad);
}

void C3dFileInfo::set_analog_samples_per_frame(int aspf)
{
	if (aspf != m_fi1.samples_per_frame)
		throw NotImplemented();
}

void C3dFileInfo::set_analog_channels(int ac)
{
	if (ac != m_fi1.analog_channels)
		throw NotImplemented();
}

std::auto_ptr<UuIcsC3d::C3dFile> C3dFileInfo::open() const
{
	return std::auto_ptr<UuIcsC3d::C3dFile>(new C3dFile(*this));
}

// **** C3dFile ****

struct UuIcsC3d::C3dFile::Impl {
	Impl(C3dFileInfo const &fi);
	FileInfo1 fi1;
	string filename;
	boost::shared_ptr<BasicIO> io;
	shared_ptr<FILE> fp;
	mutable int cached_frame_no;
	mutable FrameData cache;
	void cache_frame(int frame_no) const;
	void read_frame() const;
};

UuIcsC3d::C3dFile::Impl::Impl(C3dFileInfo const &fi)
: fi1(fi.fi1()), filename(fi.content().filename()), io(fi.content().io()), cached_frame_no(-1)
{
	fp = open_c_file(filename.c_str());
	cache.points.resize(fi1.points_per_frame);
	cache.analog_data.resize(fi1.samples_per_frame);
	for (int i=0; i<fi1.samples_per_frame; ++i) {
		cache.analog_data[i].resize(fi1.analog_channels);
	}
}

void UuIcsC3d::C3dFile::Impl::read_frame() const
{
	FILE *filep = fp.get();
	float sc = fi1.point_scale;
	if (fi1.is_integer) {
		//throw UuIcsC3d::NotImplemented();
		unsigned char buf[2];
		for (int i=0; i<fi1.points_per_frame; ++i) {
			DataPoint3d &dp = cache.points[i];
			float x, y, z;
			if (fread(buf, 1, 2, filep)!=2)
				throw ContentError("No data read", filename);
			x = sc*io->to_uint16(buf);
			if (fread(buf, 1, 2, filep)!=2)
				throw ContentError("No data read", filename);
			y = sc*io->to_uint16(buf);
			if (fread(buf, 1, 2, filep)!=2)
				throw ContentError("No data read", filename);
			z = sc*io->to_uint16(buf);
			if (fread(buf, 1, 2, filep)!=2)
				throw ContentError("No data read", filename);
			dp.set_mask(buf[0] & 0x7f);
			if (buf[0]&0x80) {
				// invalid point
				dp.invalidate();
				dp.set_residual(-1.0);
			} else {
				dp.set_residual(buf[1]*sc);
				dp.set_value(x,y,z,0,
					buf[1]==0 ? DataPoint3d::Generated : DataPoint3d::Measured);
			}
		}
		for (int i=0; i<fi1.samples_per_frame; ++i) {
			vector<float> &cur=cache.analog_data[i];
			for (int j=0; j<fi1.analog_channels; ++j) {
				if (fread(buf, 1, 2, filep)!=2)
					throw ContentError("No data read", filename);
				cur[j] = sc*io->to_uint16(buf);
			}
		}
	} else {
		unsigned char buf[4];
		for (int i=0; i<fi1.points_per_frame; ++i) {
			DataPoint3d &dp = cache.points[i];
			float x, y, z;
			if (fread(buf, 1, 4, filep)!=4)
				throw ContentError("No data read", filename);
			x = io->to_float(buf);
			if (fread(buf, 1, 4, filep)!=4)
				throw ContentError("No data read", filename);
			y = io->to_float(buf);
			if (fread(buf, 1, 4, filep)!=4)
				throw ContentError("No data read", filename);
			z = io->to_float(buf);
			if (fread(buf, 1, 4, filep)!=4)
				throw ContentError("No data read", filename);
			// TODO: 
			// Check which of the bytes of the buffer are used for encoding 
			// camera mask and residual.
			// As I read the c3d spec, it should be bytes 0 and 1,  
			// but Vicon files seem to use bytes 2 and 3.
			// There even seems to be a swap of bytes 2 and 3 from nexus version 1.5 to 1.6
			/*
			// my interpretation
			dp.camera_mask = buf[0] & 0x7f;
			if (buf[0]&0x80) {
				// invalid point
				dp.type = DataPoint3d::Invalid;
				dp.residual = -1.0;
			} else {
				dp.residual = buf[1]*sc;
				dp.type = buf[1]==0? DataPoint3d::Generated: DataPoint3d::Measured;
			}
			*/
			/*
			// Vicon interpretation 1.5
			dp.flags = (static_cast<unsigned>(buf[0])<<8) + buf[1];
			dp.camera_mask = buf[2] & 0x7f;
			if (buf[2]&0x80) {
				// invalid point
				dp.type = DataPoint3d::Invalid;
				dp.residual = -1.0;
			} else {
				dp.residual = buf[3]*sc;
				dp.type = buf[3]==0? DataPoint3d::Generated: DataPoint3d::Measured;
			}
			// End of Vicon 1.5 interpretation
			*/
			// Vicon interpretation 1.6
			unsigned flags = (static_cast<unsigned>(buf[0])<<8) + buf[1];
			dp.set_mask(buf[3] & 0x7f);
			if (buf[3]&0x80) {
				// invalid point
				dp.invalidate();
				dp.set_residual(-1.0);
			} else {
				dp.set_residual(buf[2]*sc);
				//dp.type = buf[2]==0? DataPoint3d::Generated: DataPoint3d::Measured;
				dp.set_value(x,y,z,flags,
						flags==0? DataPoint3d::Measured : DataPoint3d::Generated);
			}
			// End of Vicon 1.6 interpretation

		}
		for (int i=0; i<fi1.samples_per_frame; ++i) {
			vector<float> &cur=cache.analog_data[i];
			for (int j=0; j<fi1.analog_channels; ++j) {
				if (fread(buf, 1, 4, filep)!=4)
					throw ContentError("No data read", filename);
				cur[j] = io->to_float(buf);
			}
		}
	}
}

void UuIcsC3d::C3dFile::Impl::cache_frame(int frame_no) const
{
	if (cached_frame_no == frame_no)
		return;
	if (cached_frame_no+1!=frame_no || cached_frame_no<0) {
		// position anew
		if (fseek(fp.get(),
				fi1.data_startblock0*BlockSize+frame_no*fi1.frame_size,
				SEEK_SET)!=0)
			throw UuIcsC3d::ContentError("Cannot seek to data in frame", filename);
	}
	cached_frame_no = -1;
	read_frame();
	cached_frame_no = frame_no;
}

UuIcsC3d::C3dFile::C3dFile(C3dFileInfo const &fi)
:m_impl( new Impl(fi))
{}

UuIcsC3d::C3dFile::~C3dFile()
{}


void UuIcsC3d::C3dFile::get_frame_data(FrameData &data, int frame_no) const
{
	if (frame_no <0 || m_impl->fi1.frame_count <= frame_no)
		throw PreconditionError();
	m_impl->cache_frame(frame_no);
	data = m_impl->cache;
}

UuIcsC3d::DataPoint3d UuIcsC3d::C3dFile::get_point(int frame_no, int point_no) const
{
	if (frame_no <0 || m_impl->fi1.frame_count <= frame_no)
		throw PreconditionError();
	if (point_no <0 || m_impl->fi1.points_per_frame <= point_no)
		throw PreconditionError();
	m_impl->cache_frame(frame_no);
	return m_impl->cache.points[point_no];
}

/*
bool UuIcsC3d::write_f(std::string const &fn, Content const &content, std::vector<FrameData> const &data, EncodingType et)
{
	if (et==0) {
		et = platform_encoding();
	}
	bool can_use_low_level=!content.m_dirty;
//	if (data.size()!= content.
	// check if ParameterRep (m_data_rep) is valid
	// -has anything be written
	// -does framedata correspond to parameters (number of frames, size of frame)
	// -was data originally in floating_point format
	// -is encoding the same
}
*/

namespace {
	bool same_data_size(C3dFileInfo const &fi, vector<UuIcsC3d::FrameData> const &data)
	{
		if (data.size() != static_cast<unsigned>(fi.frame_count()))
			return false;
		if (data.size()>0) {
			if (data[0].points.size() !=
					static_cast<unsigned>(fi.points_per_frame()))
				return false;
			if (data[0].analog_data.size() !=
					static_cast<unsigned>(fi.analog_channels()))
				return false;
			else 
				if (fi.analog_channels()>0 && data[0].analog_data[0].size() !=
						static_cast<unsigned>(fi.analog_samples_per_frame()))
					return false;
		}
		return true;
	}
}

namespace {
	using namespace UuIcsC3d;

	void write_point_f(FILE *fp, DataPoint3d const &dp, BasicIO const *io, float point_scale)
	{
		unsigned char buf[4];
		if (dp.is_valid()) {
			io->encode(buf, dp.x());
			fwrite(buf,1,4,fp);
			io->encode(buf, dp.y());
			fwrite(buf,1,4,fp);
			io->encode(buf, dp.z());
			fwrite(buf,1,4,fp);
		} else {
			io->encode(buf, 0.0f);
			fwrite(buf,1,4,fp);
			fwrite(buf,1,4,fp);
			fwrite(buf,1,4,fp);
		}
		// TODO
		// Also for writing the right interpretation of the spec is important.
		// See a description of the problem at the reading part
		/*
		// My interpretation of the c3d spec
		buf[0] = dp.is_valid()? dp.camera_mask: 0x80;
		buf[1] = 0;
		if (dp.type==DataPoint3d::Measured) {
			float residual = dp.residual/point_scale;
			if (residual>0xff) {
				buf[1]=0xff;
			} else if (residual>0) {
				buf[1] = static_cast<unsigned char>(residual);
			}
		}
		buf[2] = buf[3] = 0;
		// Vicon 1.5 interpretation of the c3d spec
		buf[0] = buf[1] = 0;
		buf[0] = dp.flags&0xf>>8 & 0xff;
		buf[1] = dp.flags&0xff;
		buf[2] = dp.is_valid()? dp.camera_mask: 0x80;
		buf[3] = 0;
		if (dp.type==DataPoint3d::Measured) {
			float residual = dp.residual/point_scale;
			if (residual>0xff) {
				buf[3]=0xff;
			} else if (residual>0) {
				buf[3] = static_cast<unsigned char>(residual);
			}
		}
		// end of Vicon 1.5 interpretaion
		*/
		// Vicon 1.6 interpretation of the c3d spec
		buf[0] = buf[1] = 0;
		buf[0] = dp.flags()&0xf>>8 & 0xff;
		buf[1] = dp.flags()&0xff;
		buf[3] = dp.is_valid()? dp.mask(): 0xbf;
		buf[2] = dp.is_valid()? 0: 0x80;
		if (dp.status()==DataPoint3d::Measured) {
			float residual = dp.residual()/point_scale;
			if (residual>0xff) {
				buf[2]=0xff;
			} else if (residual>0) {
				buf[2] = static_cast<unsigned char>(residual);
			}
		}
		// end of Vicon 1.6 interpretaion
		fwrite(buf,1,4,fp);
	}

	void write_bin_f(FILE *fp, UuIcsC3d::FrameData const &fd, BasicIO const *io, int pc, int cc, int spf, float point_scale)
	{
		if (pc<0 || fd.points.size() != static_cast<unsigned>(pc))
			throw UuIcsC3d::PreconditionError();
		if (cc<0 || fd.analog_data.size()!= static_cast<unsigned>(cc))
			throw UuIcsC3d::PreconditionError();
		if (cc>0 &&
				(spf<0 || fd.analog_data[0].size()!= static_cast<unsigned>(spf)))
			throw UuIcsC3d::PreconditionError();
		for (int i=0; i<pc; ++i) {
			write_point_f(fp, fd.points[i], io, point_scale);
		}
		unsigned char buf[4];
		for (int i=0; i<cc; ++i) {
			vector<float> const &analog(fd.analog_data[i]);
			for (int j=0; j<spf; ++j) {
				io->encode(buf, analog[j]);
				fwrite(buf,1,4,fp);
			}
		}
	}

}

bool UuIcsC3d::write_with_same_header(std::string const &fn,
				C3dFileInfo const &fi, std::vector<FrameData> const &data)
{
	return write(fn, fi, data, fi.content().io().get());
}

namespace {
	int compute_parameter_section_size(C3dFileInfo const &fi)
	{
		int size = 4; // 4 bytes parameter heading
		vector<Group> const & groups = fi.content().the_groups();
		for (vector<Group>::const_iterator cur = groups.begin();
						cur != groups.end(); ++cur) {
			size += cur->total_representation_size();
		}
		return size;
	}

	void write_parameter(FILE *fp, Parameter const &param, BasicIO const* encoder, string const &filename)
	{
		signed char sch;
		int sz = param.name().size();
		if (sz>127)
			throw ContentError("Parameter name too large", filename);
		sch = param.is_locked() ? -sz : sz;
		fwrite(&sch, 1, 1, fp);
		if (param.id() > 127)
			throw ContentError("Parameter id too large", filename);
		sch = param.id();
		fwrite(&sch, 1, 1, fp);
		fwrite(param.name().c_str(), 1, sz, fp);
		short desc_sz = param.representation_size()-sz-2;
		unsigned char buf[2];
		encoder->encode(buf, desc_sz);
		fwrite(buf, 1, 2, fp);
		sch = param.data_type();
		fwrite(&sch, 1, 1, fp);
        if (param.bounds().size()>7u)
            throw ContentError(
					"Parameter with more than 7 dimensions encountered",
					filename);
		sch = param.bounds().size();
		fwrite(&sch, 1, 1, fp);
		for (int d=param.bounds().size(); d>0; --d) {
			// In file the Fortran array convention is used:
			//(fastest changing index in memory first)
			// In our library we use the C convention
			//(fastest changing in memory last)
			unsigned char ch = param.bounds()[d-1];
			fwrite(&ch, 1, 1, fp);
		}
		if (!param.raw_data().empty())
			fwrite(&(param.raw_data()[0]), 1, param.raw_data().size(), fp);
		sz = param.description().size();
		if (sz>127)
			throw ContentError("Description of parameter too large", filename);
		sch = sz;
		fwrite(&sch, 1, 1, fp);
		fwrite(param.description().c_str(), 1, sz, fp);
	}

	void write_group_data(FILE *fp, Group const &group, 
		BasicIO const* encoder, string const &filename)
	{
		signed char sch;
		int sz = group.name().size();
		if (sz>127)
			throw ContentError("Group name too large", filename);
		sch = group.is_locked() ? -sz : sz;
		fwrite(&sch, 1, 1, fp);
		if (group.id() > 127)
			throw ContentError("Group id too large", filename);
		sch = -group.id();
		fwrite(&sch, 1, 1, fp);
		fwrite(group.name().c_str(), 1, sz, fp);
		sz = group.description().size();
		if (sz>127)
			throw ContentError("Description of group too large", filename);
		unsigned char buf[2];
		encoder->encode(buf, static_cast<short>(sz+3));
		fwrite(buf, 1, 2, fp);
		sch = sz;
		fwrite(&sch, 1, 1, fp);
		fwrite(group.description().c_str(), 1, sz, fp);
	}


	void write_parameters_of_group(FILE *fp, Group const &group, 
		BasicIO const* encoder, string const &filename)
	{
		for (vector<Parameter>::const_iterator p=group.the_parameters().begin();
			p != group.the_parameters().end(); ++p) {
				write_parameter(fp, *p, encoder, filename);
		}
	}

	void write_group_with_parameters(FILE *fp, Group const &group, 
		BasicIO const* encoder, string const &filename)
	{
		write_group_data(fp, group, encoder, filename);
		write_parameters_of_group(fp, group, encoder, filename);
	}

}

bool UuIcsC3d::write(std::string const &fn, C3dFileInfo fi,
				std::vector<FrameData> const &data, BasicIO const* encoder)
{
	if (encoder->get_encoding()!=fi.content().io()->get_encoding()) {
		throw NotImplemented();
	}
	string filename = fi.content().filename();
	// write header
	// count parameter section size

	int frame_count = data.size();
	int point_count = 0;
	int channel_count = 0;
	int samples_per_frame = 0;
	if (frame_count>0) {
		point_count = data[0].points.size();
		channel_count = data[0].analog_data.size();
		if (channel_count>0)
			samples_per_frame = data[0].analog_data[0].size();
	}
	fi.set_frame_count(frame_count);
	fi.set_points_per_frame(point_count);
	fi.set_analog_channels(channel_count);
	fi.set_analog_samples_per_frame(samples_per_frame);
	int param_section_size = compute_parameter_section_size(fi);
	int block_count = (param_section_size-1)/BlockSize +1;
	if (block_count>255)
		throw ContentError("Parameter section too large",filename);
	fi.set_param_blocks(block_count);

	// write header section
	UuIcsC3d::Block header_block;
	write_header_block(header_block, fi.content().header(), encoder);
	shared_ptr<FILE> fp(open_c_file(fn.c_str(), "wb"));
	fwrite(header_block, 1, BlockSize, fp.get());
	
	// write parameter section

	unsigned char ch=1;
	fwrite(&ch, 1, 1,fp.get());
	ch = 0x50;
	fwrite(&ch, 1, 1,fp.get());
	ch = block_count;
	fwrite(&ch, 1, 1,fp.get());
	ch = encoder->get_encoding();
	fwrite(&ch, 1, 1,fp.get());
	// write groups and parameters
	vector<Group> const & groups = fi.content().the_groups();
#if 1
	for (vector<Group>::const_iterator cur = groups.begin();
					cur != groups.end(); ++cur) {
		write_group_with_parameters(fp.get(), *cur, encoder, filename);
	}
#else
	for (vector<Group>::const_iterator cur = groups.begin();
					cur != groups.end(); ++cur) {
		write_group_data(fp.get(), *cur, encoder, filename);
	}
	for (vector<Group>::const_iterator cur = groups.begin();
					cur != groups.end(); ++cur) {
		write_parameters_of_group(fp.get(), *cur, encoder, filename);
	}
#endif
	// Pad with 0
	ch=0;
	for(int pad=block_count*BlockSize-param_section_size; pad>0; --pad)
		fwrite(&ch, 1, 1,fp.get());

	// write point and analog data
	if (frame_count>0) {
		for (int i=0; i<frame_count; ++i) {
			write_bin_f(fp.get(), data[i], encoder, point_count, channel_count,
							samples_per_frame, fi.fi1().point_scale);
		}
	}
	return true;
}




