#include "skeleton_info.hpp"
#include "uuc3d.hpp"
#include <iostream>
#include <fstream>
#include <algorithm>
#include <math.h>
#include "subjects.hpp"

using std::string;
using std::vector;
using std::cerr;
using UuIcsC3d::SpacePaddedString;
using UuIcsC3d::FrameData;
using UuIcsC3d::DataPoint3d;
using UuIcsC3d::LabeledSkeleton;

UuIcsC3d::SkeletonInfo const &get_uu_skeleton_info();

namespace {
	struct GroupElement {
		int group;
		int element;
	};

/*
	void check_skeleton_info()
	{
		// visual consistency check for skeleton
		using std::cout;
		UuIcsC3d::SkeletonInfo const &skinf = get_uu_skeleton_info();
		cout<<skinf.group_count()<<" groups.\n";
		for (int i=0; i<skinf.group_count(); ++i) {
			cout<<"Group "<<skinf.group_name(i)<<"\n";
			UuIcsC3d::ConstNameRange members(skinf.group_members(i));
			for (std::string const *cur =members.begin(); cur !=members.end(); ++cur) {
				cout<<"\t"<<*cur<<"\n";
			}
			vector<int> cnx = skinf.group_connections(i);
			cout<<"  connections: \n";
			for (unsigned j=0; j<cnx.size(); ++j) {
				cout<<"\t"<<skinf.group_name(cnx[j])<<"\n";
			}
		}
	}


	inline int triangle_count(int n)
	{
		return n*(n-1)/2;
	}
*/

	bool is_subject_label(int label, vector<unsigned int> const &no_label)
	{
		return std::find(no_label.begin(), no_label.end(), label) == no_label.end();
	}


	//! \brief Pair of labels for a single subject
	struct MarkerPair {
		int subject;
		GroupElement grel[2];
		int point_index[2];  // index in specific recording
	};

	//! \brief Distances for a pair of markers in different frames
	struct MarkerDists {
		MarkerPair mp;
		vector<float> sqdist;
	};

	//! \brief Contains a frame number and a distance.
	struct FrameDist {
		int frame;
		float dist;
	};

	struct DistanceStat {
		static const int PercentCount = 7;
		static float percent_values[PercentCount];
		static const int FixedDistCount = 4;
		static float delta[FixedDistCount];
		MarkerPair mp;
		FrameDist percent_dist[PercentCount]; // dists at 0, 5 10, 50 (median), 90 95 and 100 %
		float valid_fraction;
		// fraction of distances that are valid (both markers valid in this frame)
		float delta_fraction[FixedDistCount]; 
		// fraction of distances within 1,2,3,4 mm from median distance
		// invalid points are counted too, so can never be more than valid_fraction.
	};

	float DistanceStat::percent_values[PercentCount] = {0.0f, 0.05f, 0.1f, 0.5f, 0.9f, 0.95f, 1.0f};
	float DistanceStat::delta[FixedDistCount] = {1.0f, 2.0f, 3.0f, 4.0f};

	//! \brief Contains a frame number and the number of missing markers
	struct FrameMiss {
		FrameMiss() : frame(-1), missing_count(0) {}
		FrameMiss(int f, int m): frame(f), missing_count(m) {}
		int frame;
		int missing_count;
	};

	struct DisplayOrder {
		bool operator()(FrameMiss const &fm1, FrameMiss const &fm2)
		{
			if (fm1.missing_count > fm2.missing_count)
				return true;
			if (fm1.missing_count < fm2.missing_count)
				return false;
			return (fm1.frame < fm2.frame);
		}
	};

	//! \brief Used to collect information about missing markers in a frame. 
	struct MissingCandidate {
		MissingCandidate() :  high_in_range(false) {}
		//! The number of missing markers in a frame
		FrameMiss fm;
		//! Whether the number of missing markers is higher than in the surrounding N frames
		bool high_in_range;
	};

	//! \brief Class that collects statistics about a c3dfile
	//
	struct LabellingStatistics {
		LabellingStatistics(int ppf, vector<LabeledSkeleton> const&subjects, vector<unsigned> const &no_label);
		void next_frame(int frame, FrameData const &fd);
		vector<int> invalid_count() const {return m_count;}
		vector<DistanceStat> distance_statistics() const;
		vector<FrameMiss> missing_label_maxima() const {return m_missing_label_max;}
	private:
		vector<LabeledSkeleton> m_subjects;
		vector<int> m_count;
		vector<unsigned> m_no_label;
		vector<MarkerDists> m_dists;
		static int const MisMaxN = 9;
		int m_mis_lab_idx;
		MissingCandidate m_mis_lab[MisMaxN];
		vector<FrameMiss> m_missing_label_max;
	};


	LabellingStatistics::LabellingStatistics(int ppf, vector<LabeledSkeleton> const&subjects, vector<unsigned> const &no_label)
		: m_subjects(subjects), m_count(ppf,0), m_no_label(no_label), m_mis_lab_idx(0)
	{
		// compute which distance pairs should be computed
		MarkerDists md;
		MarkerPair &mp = md.mp;
		for (size_t subject=0; subject<m_subjects.size(); ++subject) {
			mp.subject = subject;
			vector<vector<int> > const &point_index=m_subjects[subject].point_index();
			for (size_t group=0; group<point_index.size(); ++group) {
				mp.grel[0].group = group;
				mp.grel[1].group = group;
				vector<int> const &members=point_index[group];
				size_t sz = members.size();
				for (size_t el1=0; el1+1<sz; ++el1) {
					mp.grel[0].element = el1;
					mp.point_index[0] = members[el1];
					for (size_t el2=el1+1; el2<sz; ++el2) {
						mp.grel[1].element = el2;
						mp.point_index[1] = members[el2];
						if (mp.point_index[0]>=0 && mp.point_index[1]>=0)
							m_dists.push_back(md);
					}
				}
			}
		}
	}

	float sqdistance(DataPoint3d const &dp1, DataPoint3d const &dp2)
	{
		if (!dp1.is_valid() || !dp2.is_valid())
			return -1.0f;
		double sqdist = 0.0;
		for (int i=0; i<3; ++i) {
			double d = dp1.coordinates()[i]-dp2.coordinates()[i];
			sqdist += d*d;
		}
		return static_cast<float>(sqdist);
	}


	void LabellingStatistics::next_frame(int frame, FrameData const &frame_data)
	{
		int missing_count = 0;
		for (size_t j=0; j<m_count.size(); ++j) {
			DataPoint3d const &dp=frame_data.points[j];
			if (!dp.is_valid()) {
				++m_count[j];
				if (is_subject_label(j, m_no_label))
					++missing_count;
			}
		}
		++m_mis_lab_idx;
		if (m_mis_lab_idx == MisMaxN)
			m_mis_lab_idx=0;
		bool high_in_range=true;
		for (int i=0; i<MisMaxN; ++i) {
			if (m_mis_lab[i].fm.missing_count >= missing_count) {
				high_in_range = false;
			} else {
				m_mis_lab[i].high_in_range = false;
			}
		}
		if (m_mis_lab[m_mis_lab_idx].high_in_range) {
			m_missing_label_max.push_back(m_mis_lab[m_mis_lab_idx].fm);
		}
		m_mis_lab[m_mis_lab_idx].fm = FrameMiss(frame, missing_count);
		m_mis_lab[m_mis_lab_idx].high_in_range = high_in_range;

		for (size_t d=0; d<m_dists.size(); ++d) {
			MarkerDists &md=m_dists[d];
			md.sqdist.push_back(
				sqdistance(frame_data.points[md.mp.point_index[0]],frame_data.points[md.mp.point_index[1]]));
		}
	}

	struct IndirectLessByRef {
		IndirectLessByRef(vector<float> const &values) :m_values(values) {}
		bool operator()(int i1, int i2)
		{
			return m_values[i1] < m_values[i2];
		}
	private:
		vector<float> const &m_values;
	};

	struct IndirGeqThan {
		IndirGeqThan(float val, vector<float> const &values):m_val(val), m_values(values) {}
		bool operator()(int i)
		{
			return m_values[i]>=m_val;
		}
	private:
		float m_val;
		vector<float> const &m_values;
	};

	vector<int> sorted_index(vector<float> const &values)
	{
		vector<int> result(values.size());
		for (size_t i=0; i<result.size(); ++i)
			result[i]=i;
		std::sort(result.begin(), result.end(), IndirectLessByRef(values));
		return result;
	}

	int relative_index(int size, double frac)
	{
		return static_cast<int>((size-1)*frac+0.5);
	}

	vector<DistanceStat> LabellingStatistics::distance_statistics() const
	{
		vector<DistanceStat> result;
		result.reserve(m_dists.size());
		for (size_t i=0; i<m_dists.size(); ++i) {
			vector<float> const &sqdist = m_dists[i].sqdist;
			vector<int> si = sorted_index(sqdist);
			vector<int>::iterator cur;
			cur=find_if(si.begin(), si.end(),IndirGeqThan(0.0, sqdist));
			size_t sz = std::distance(cur, si.end());
			DistanceStat ds;
			ds.mp = m_dists[i].mp;
			if (sz==0) {
				// degenerate case: no valid distances.
				ds.valid_fraction = 0.0;
				FrameDist fd;
				fd.dist=-1.0f;
				fd.frame=-1;
				for (int j=0; j<DistanceStat::PercentCount; ++j) {
					ds.percent_dist[j] = fd;
				}
				for (int j=0; j<DistanceStat::FixedDistCount; ++j) {
					ds.delta_fraction[j] = 0.0;
				}
			} else {
				FrameDist fd;
				for (int j=0; j<DistanceStat::PercentCount; ++j) {
					fd.frame = *(cur+relative_index(sz, DistanceStat::percent_values[j]));
					fd.dist = sqrt(sqdist[fd.frame]);
					ds.percent_dist[j] = fd;
				}
				float median = ds.percent_dist[DistanceStat::PercentCount/2].dist;
				cur=find_if(si.begin(), si.end(),IndirGeqThan(0.0, sqdist));
				int invalid_count = std::distance(si.begin(), cur);
				int before_count[DistanceStat::FixedDistCount], after_count[DistanceStat::FixedDistCount];
				for (int j=DistanceStat::FixedDistCount; j>0; --j) {
					float cut_off = median - DistanceStat::delta[j-1];
					if (cut_off<0.0) cut_off = 0.0;
					cur=find_if(cur, si.end(),IndirGeqThan(cut_off*cut_off, sqdist));
					before_count[j-1] = std::distance(si.begin(), cur);
				}
				for (int j=0; j<DistanceStat::FixedDistCount; ++j) {
					float cut_off = median + DistanceStat::delta[j];
					cur=find_if(cur, si.end(),IndirGeqThan(cut_off*cut_off, sqdist));
					after_count[j] = std::distance(cur, si.end());
				}
				int total = std::distance(si.begin(), si.end());
				ds.valid_fraction = (total-invalid_count)*1.0f/total;
				for (int j=0; j<DistanceStat::FixedDistCount; ++j) {
					int inside = total - before_count[j] - after_count[j];
					ds.delta_fraction[j] = inside*1.0f/total;
				}
			}
			result.push_back(ds);
		}
		return result;
	}

	struct LabelInvalid { string label; int invalid;};

	struct GreaterByInvalid {
		bool operator()(LabelInvalid const &li1, LabelInvalid const &li2)
		{
			return li1.invalid>li2.invalid;
		}
	};

	void do_work(string infile, string statfile)
	{
		//	check_skeleton_info();
		//	check_basis();
		UuIcsC3d::C3dFileInfo fi(infile);
		int ppf = fi.points_per_frame(); 
		if (ppf <= 0)
			return;
		int fc = fi.frame_count();
		vector<SpacePaddedString> labels(fi.point_labels());
		vector<LabeledSkeleton> subjects;
		vector<unsigned> no_label = fill_skeletons(
                    subjects, labels, get_uu_skeleton_info());
		for (int i=labels.size(); i<ppf; ++i) {
			labels.push_back(SpacePaddedString("<no label>"));
		}
		LabellingStatistics stats(ppf, subjects, no_label);
		std::auto_ptr<UuIcsC3d::C3dFile> filep(fi.open());
		vector<FrameData> frame_data(fc);
		for (int i=0; i<fc; ++i) {
			filep->get_frame_data(frame_data[i], i);
			stats.next_frame(i, frame_data[i]);
		}
		filep.reset();

		// report
		std::ofstream statfs(statfile.c_str());
		unsigned detected_subject_labels = ppf-no_label.size();
		size_t subject_count = subjects.size();
		unsigned expected_subject_labels = subject_count==0 ? 0 : subject_count*subjects[0].label_count();
		vector<LabelInvalid> invalid_count(labels.size());
		{
			vector<int> const &iv(stats.invalid_count());
			for (size_t l=0; l!= iv.size(); ++l) {
				invalid_count[l].label = labels[l].stripped();
				invalid_count[l].invalid = iv[l];
			}
		}
		int missing_total=(expected_subject_labels-detected_subject_labels)*fc;
		int unclassified_total=0;
		for (size_t i=0; i<invalid_count.size(); ++i) {
			if (is_subject_label(i, no_label)) {
				missing_total += invalid_count[i].invalid;
			} else {
				unclassified_total += (fc-invalid_count[i].invalid);
			}
		}
		statfs<<"name\tframes\texpected labels\tdetected labels\tmissing\tmis_fm\t%mis\tanony_max\tanony\tanony_fm\n";
		statfs<<infile<<"\t"<<fc<<"\t"<<expected_subject_labels<<"\t"<<detected_subject_labels
			<<"\t"<<missing_total<<"\t"<<(1.0*missing_total/fc)<<"\t"<<(100.0*missing_total/fc/expected_subject_labels)
			<<"\t"<<no_label.size()<<"\t"<<unclassified_total<<"\t"<<(1.0*unclassified_total/fc)<<"\n";
		statfs<<"*** Marker statistics ***\n";
		statfs<<"marker name\tinvalid\n";
		if (expected_subject_labels != detected_subject_labels) {
			for (size_t s=0; s!=subjects.size(); ++s) {
				for (size_t g =0; g != subjects[s].point_index().size(); ++g) {
					vector<int> const &pi = subjects[s].point_index()[g];
					for (size_t m=0; m!=pi.size(); ++m) {
						if (pi[m] < 0) {
							statfs<<subjects[g].name()<<":"<<get_uu_skeleton_info().group_members(g).begin()+m<<"\t"<<fc<<"\n";
						}
					}
				}
			}
		}
		sort(invalid_count.begin(), invalid_count.end(), GreaterByInvalid());
		for (size_t i=0; i<invalid_count.size(); ++i) {
			statfs<<invalid_count[i].label<<"\t"<<invalid_count[i].invalid<<"\n";
		}

		vector<FrameMiss> missing_label_maxima = stats.missing_label_maxima();
		sort(missing_label_maxima.begin(), missing_label_maxima.end(), DisplayOrder());
		statfs<<"*** Maxima of missing labels ***\n"
			<<"frame\tcount\n";
		for (size_t i=0; i!= missing_label_maxima.size(); ++i) {
			statfs<<missing_label_maxima[i].frame<<'\t'<<missing_label_maxima[i].missing_count<<"\n";
		}

		vector<DistanceStat> const &ds(stats.distance_statistics());
		statfs<<"*** Distance statistics ***\n";
		statfs<<"Subject\tmarker1\tmarker2\tvalid fraction";
		for (int j=0; j<DistanceStat::FixedDistCount; ++j) {
			float delta = DistanceStat::delta[j];
			statfs<<"\tinside "<<delta;
		}
		statfs<<"\tsmallest\tfrm\t5%\tfrm\t10%\tfrm\tmedian\tfrm\t90%\tfrm\t95\tfrm%\tlargest\tfrm\n";
		for (size_t i=0; i<ds.size(); ++i) {
			DistanceStat const &cur = ds[i];
			statfs<< cur.mp.subject<<"\t"<<labels[cur.mp.point_index[0]].stripped()<<"\t"<<labels[cur.mp.point_index[1]].stripped();
			statfs<<"\t"<<cur.valid_fraction;
			for (int j=0; j<DistanceStat::FixedDistCount; ++j) {
				statfs<<"\t"<<cur.delta_fraction[j];
			}
			for (int j=0; j<DistanceStat::PercentCount;++j) {
				statfs<<"\t"<<cur.percent_dist[j].dist<<"\t"<<cur.percent_dist[j].frame;
			}
			statfs<<"\n";
		}

	}
}

char *progname=0;

void print_usage(std::ostream &os)
{
	os<<"Usage:\n";
	os<<progname<<" input_c3d statistics_file\n";
}

int main(int argc, char* argv[])
{
	progname = argv[0];
	if (argc<3) {
		print_usage(cerr);
		return 1;
	}
	try {
		do_work(argv[1], argv[2]);
	} catch (UuIcsC3d::OpenError const &err) {
		cerr<<"An error occurred while opening file "<<err.filename();
		return 2;
	}
	return 0;
}
